@@ -1,8 +1,9 @@
 from __future__ import annotations
 
 import io
-from typing import Any
+from weakref import WeakValueDictionary
 
+from requests import ReadTimeout
 from sapiopylib.rest.ClientCallbackService import ClientCallback
 from sapiopylib.rest.DataMgmtService import DataMgmtServer
 from sapiopylib.rest.User import SapioUser
@@ -15,18 +16,20 @@ from sapiopylib.rest.pojo.datatype.FieldDefinition import AbstractVeloxFieldDefi
 from sapiopylib.rest.pojo.webhook.ClientCallbackRequest import OptionDialogRequest, ListDialogRequest, \
     FormEntryDialogRequest, InputDialogCriteria, TableEntryDialogRequest, ESigningRequestPojo, \
     DataRecordDialogRequest, InputSelectionRequest, FilePromptRequest, MultiFilePromptRequest, \
-    TempTableSelectionRequest
+    TempTableSelectionRequest, DisplayPopupRequest, PopupType
 from sapiopylib.rest.pojo.webhook.ClientCallbackResult import ESigningResponsePojo
-from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
 from sapiopylib.rest.pojo.webhook.WebhookEnums import FormAccessLevel, ScanToSelectCriteria, SearchType
 from sapiopylib.rest.utils.DataTypeCacheManager import DataTypeCacheManager
 from sapiopylib.rest.utils.FormBuilder import FormBuilder
 from sapiopylib.rest.utils.recorddatasinks import InMemoryRecordDataSink
 from sapiopylib.rest.utils.recordmodel.RecordModelWrapper import WrappedType
 
-from sapiopycommons.general.aliases import FieldMap, SapioRecord, AliasUtil, RecordIdentifier
+from sapiopycommons.files.file_util import FileUtil
+from sapiopycommons.general.aliases import FieldMap, SapioRecord, AliasUtil, RecordIdentifier, FieldValue, \
+    UserIdentifier
 from sapiopycommons.general.custom_report_util import CustomReportUtil
-from sapiopycommons.general.exceptions import SapioUserCancelledException, SapioException, SapioUserErrorException
+from sapiopycommons.general.exceptions import SapioUserCancelledException, SapioException, SapioUserErrorException, \
+    SapioDialogTimeoutException
 from sapiopycommons.recordmodel.record_handler import RecordHandler
 
 
@@ -34,29 +37,104 @@ class CallbackUtil:
     user: SapioUser
     callback: ClientCallback
     dt_cache: DataTypeCacheManager
+    _original_timeout: int
+    timeout_seconds: int
     width_pixels: int | None
     width_percent: float | None
 
-    def __init__(self, context: SapioWebhookContext | SapioUser):
+    __instances: WeakValueDictionary[SapioUser, CallbackUtil] = WeakValueDictionary()
+    __initialized: bool
+
+    def __new__(cls, context: UserIdentifier):
         """
         :param context: The current webhook context or a user object to send requests from.
         """
-        self.user = context if isinstance(context, SapioUser) else context.user
+        user = AliasUtil.to_sapio_user(context)
+        obj = cls.__instances.get(user)
+        if not obj:
+            obj = object.__new__(cls)
+            obj.__initialized = False
+            cls.__instances[user] = obj
+        return obj
+
+    def __init__(self, context: UserIdentifier):
+        """
+        :param context: The current webhook context or a user object to send requests from.
+        """
+        if self.__initialized:
+            return
+        self.__initialized = True
+
+        self.user = AliasUtil.to_sapio_user(context)
         self.callback = DataMgmtServer.get_client_callback(self.user)
         self.dt_cache = DataTypeCacheManager(self.user)
+        self._original_timeout = self.user.timeout_seconds
+        self.timeout_seconds = self.user.timeout_seconds
         self.width_pixels = None
         self.width_percent = None
 
-    def set_dialog_width(self, width_pixels: int | None, width_percent: float | None):
+    def set_dialog_width(self, width_pixels: int | None = None, width_percent: float | None = None):
         """
         Set the width that dialogs will appear as for those dialogs that support specifying their width.
 
         :param width_pixels: The number of pixels wide that dialogs will appear as.
-        :param width_percent: The percentage of the client's screen width that dialogs will appear as.
+        :param width_percent: The percentage (as a value between 0 and 1) of the client's screen width that dialogs
+            will appear as.
         """
+        if width_pixels is not None and width_percent is not None:
+            raise SapioException("Cannot set both width_pixels and width_percent at once.")
         self.width_pixels = width_pixels
         self.width_percent = width_percent
-    
+
+    def set_dialog_timeout(self, timeout: int):
+        """
+        Alter the timeout time used for callback requests that create dialogs for the user to interact with. By default,
+        a CallbackUtil will use the timeout time of the SapioUser provided to it. By altering this, a different timeout
+        time is used.
+
+        :param timeout: The number of seconds that must elapse before a SapioDialogTimeoutException is thrown by
+            any callback that creates a dialog for the user to interact with.
+        """
+        self.timeout_seconds = timeout
+
+    def toaster_popup(self, message: str, title: str = "", popup_type: PopupType = PopupType.Info) -> None:
+        """
+        Display a toaster popup in the bottom right corner of the user's screen.
+
+        :param message: The message to display in the toaster.
+        :param title: The title to display at the top of the toaster.
+        :param popup_type: The popup type to use for the toaster. This controls the color that the toaster appears with.
+            Info is blue, Success is green, Warning is yellow, and Error is red
+        """
+        self.callback.display_popup(DisplayPopupRequest(title, message, popup_type))
+
+    def display_info(self, message: str) -> None:
+        """
+        Display an info message to the user in a dialog. Repeated calls to this function will append the new messages
+        to the same dialog if it is still opened by the user.
+
+        :param message: The message to display to the user.
+        """
+        self.callback.display_info(message)
+
+    def display_warning(self, message: str) -> None:
+        """
+        Display a warning message to the user in a dialog. Repeated calls to this function will append the new messages
+        to the same dialog if it is still opened by the user.
+
+        :param message: The message to display to the user.
+        """
+        self.callback.display_warning(message)
+
+    def display_error(self, message: str) -> None:
+        """
+        Display an error message to the user in a dialog. Repeated calls to this function will append the new messages
+        to the same dialog if it is still opened by the user.
+
+        :param message: The message to display to the user.
+        """
+        self.callback.display_error(message)
+
     def option_dialog(self, title: str, msg: str, options: list[str], default_option: int = 0,
                       user_can_cancel: bool = False) -> str:
         """
@@ -71,8 +149,15 @@ class CallbackUtil:
             SapioUserCancelledException is thrown.
         :return: The name of the button that the user selected.
         """
-        request = OptionDialogRequest(title, msg, options, default_option, user_can_cancel)
-        response: int | None = self.callback.show_option_dialog(request)
+        request = OptionDialogRequest(title, msg, options, default_option, user_can_cancel,
+                                      width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: int | None = self.callback.show_option_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return options[response]
@@ -111,17 +196,27 @@ class CallbackUtil:
         """
         return self.option_dialog(title, msg, ["Yes", "No"], 0 if default_yes else 1, False) == "Yes"
 
-    def list_dialog(self, title: str, options: list[str], multi_select: bool = False) -> list[str]:
+    def list_dialog(self, title: str, options: list[str], multi_select: bool = False,
+                    preselected_values: list[str] | None = None) -> list[str]:
         """
         Create a list dialog with the given options for the user to choose from.
 
         :param title: The title of the dialog.
         :param options: The list options that the user has to choose from.
         :param multi_select: Whether the user is able to select multiple options from the list.
+        :param preselected_values: A list of values that will already be selected when the list dialog is created. The
+            user can unselect these values if they want to.
         :return: The list of options that the user selected.
         """
-        request = ListDialogRequest(title, multi_select, options)
-        response: list[str] | None = self.callback.show_list_dialog(request)
+        request = ListDialogRequest(title, multi_select, options, preselected_values,
+                                    width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[str] | None = self.callback.show_list_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -163,8 +258,6 @@ class CallbackUtil:
         builder = FormBuilder(data_type, display_name, plural_display_name)
         for field_def in fields:
             field_name = field_def.data_field_name
-            if values and hasattr(field_def, "default_value"):
-                field_def.default_value = values.get(field_name)
             column: int = 0
             span: int = 4
             if column_positions and field_name in column_positions:
@@ -173,8 +266,15 @@ class CallbackUtil:
                 span = position[1]
             builder.add_field(field_def, column, span)
 
-        request = FormEntryDialogRequest(title, msg, builder.get_temporary_data_type())
-        response: FieldMap | None = self.callback.show_form_entry_dialog(request)
+        request = FormEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values,
+                                         width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: FieldMap | None = self.callback.show_form_entry_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -215,13 +315,13 @@ class CallbackUtil:
         modifier = FieldModifier(visible=True, editable=editable)
 
         # Build the form using only those fields that are desired.
+        values: dict[str, FieldValue] = {}
         builder = FormBuilder(data_type, type_def.display_name, type_def.plural_display_name)
         for field_name in fields:
             field_def = field_defs.get(field_name)
             if field_def is None:
                 raise SapioException(f"No field of name \"{field_name}\" in field definitions of type \"{data_type}\"")
-            if hasattr(field_def, "default_value"):
-                field_def.default_value = record.get_field_value(field_name)
+            values[field_name] = record.get_field_value(field_name)
             column: int = 0
             span: int = 4
             if column_positions and field_name in column_positions:
@@ -230,13 +330,20 @@ class CallbackUtil:
                 span = position[1]
             builder.add_field(modifier.modify_field(field_def), column, span)
 
-        request = FormEntryDialogRequest(title, msg, builder.get_temporary_data_type())
-        response: FieldMap | None = self.callback.show_form_entry_dialog(request)
+        request = FormEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values,
+                                         width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: FieldMap | None = self.callback.show_form_entry_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
 
-    def input_dialog(self, title: str, msg: str, field: AbstractVeloxFieldDefinition) -> Any:
+    def input_dialog(self, title: str, msg: str, field: AbstractVeloxFieldDefinition) -> FieldValue:
         """
         Create an input dialog where the user must input data for a singular field.
 
@@ -245,8 +352,15 @@ class CallbackUtil:
         :param field: The definition for a field that the user must provide input to.
         :return: The response value from the user for the given field.
         """
-        request = InputDialogCriteria(title, msg, field, self.width_pixels, self.width_percent)
-        response: Any | None = self.callback.show_input_dialog(request)
+        request = InputDialogCriteria(title, msg, field,
+                                      width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: FieldValue | None = self.callback.show_input_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -322,6 +436,8 @@ class CallbackUtil:
                      msg: str,
                      fields: list[AbstractVeloxFieldDefinition],
                      values: list[FieldMap],
+                     group_by: str | None = None,
+                     image_data: list[bytes] | None = None,
                      *,
                      data_type: str = "Default",
                      display_name: str | None = None,
@@ -335,6 +451,10 @@ class CallbackUtil:
         :param fields: The definitions of the fields to display as table columns. Fields will be displayed in the order
             they are provided in this list.
         :param values: The values to set for each row of the table.
+        :param group_by: If provided, the created table dialog will be grouped by the field with this name by default.
+            The user may remove this grouping if they want to.
+        :param image_data: The bytes to the images that should be displayed in the rows of the table. Each element in
+            the image data list corresponds to the element at the same index in the values list.
         :param data_type: The data type name for the temporary data type that will be created for this table.
         :param display_name: The display name for the temporary data type. If not provided, defaults to the data type
             name.
@@ -356,8 +476,16 @@ class CallbackUtil:
         for field in fields:
             builder.add_field(modifier.modify_field(field))
 
-        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values)
-        response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values,
+                                          record_image_data_list=image_data, group_by_field=group_by,
+                                          width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -367,7 +495,9 @@ class CallbackUtil:
                             msg: str,
                             fields: list[str],
                             records: list[SapioRecord],
-                            editable: bool | None = True) -> list[FieldMap]:
+                            editable: bool | None = True,
+                            group_by: str | None = None,
+                            image_data: list[bytes] | None = None) -> list[FieldMap]:
         """
         Create a table dialog where the user may input data into the fields of the table. The table is constructed from
         a given list of records of a singular type. Provided field names must match fields on the definition of the data
@@ -384,9 +514,15 @@ class CallbackUtil:
             they are provided in this list.
         :param editable: If true, all fields are displayed as editable. If false, all fields are displayed as
             uneditable. If none, only those fields that are defined as editable by the data designer will be editable.
+        :param group_by: If provided, the created table dialog will be grouped by the field with this name by default.
+            The user may remove this grouping if they want to.
+        :param image_data: The bytes to the images that should be displayed in the rows of the table. Each element in
+            the image data list corresponds to the element at the same index in the records list.
         :return: A list of dictionaries mapping the data field names of the given field definitions to the response
             value from the user for that field for each row.
         """
+        if not records:
+            raise SapioException("No records provided.")
         data_types: set[str] = {x.data_type_name for x in records}
         if len(data_types) > 1:
             raise SapioException("Multiple data type names encountered in records list for record table popup.")
@@ -410,8 +546,16 @@ class CallbackUtil:
                 raise SapioException(f"No field of name \"{field_name}\" in field definitions of type \"{data_type}\"")
             builder.add_field(modifier.modify_field(field_def))
 
-        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), field_map_list)
-        response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), field_map_list,
+                                          record_image_data_list=image_data, group_by_field=group_by,
+                                          width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -522,10 +666,10 @@ class CallbackUtil:
             field_names.append(name)
 
         # Get the values for each row.
-        values: list[dict[str, Any]] = []
+        values: list[dict[str, FieldValue]] = []
         for row in row_contents:
             # The final values for this row:
-            row_values: dict[str, Any] = {}
+            row_values: dict[str, FieldValue] = {}
 
             # Map the records for this row by their data type. If a field map is provided, its data type is Default.
             row_records: dict[str, SapioRecord | FieldMap] = {}
@@ -567,8 +711,15 @@ class CallbackUtil:
         for field in final_fields:
             builder.add_field(field)
 
-        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values)
-        response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        request = TableEntryDialogRequest(title, msg, builder.get_temporary_data_type(), values,
+                                          width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[FieldMap] | None = self.callback.show_table_entry_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
@@ -616,8 +767,15 @@ class CallbackUtil:
                 raise SapioException(f"The data type \"{data_type}\" does not have a layout by the name "
                                      f"\"{layout_name}\" in the system.")
 
-        request = DataRecordDialogRequest(title, record, layout, minimized, access_level, plugin_path_list)
-        response: bool = self.callback.data_record_form_view_dialog(request)
+        request = DataRecordDialogRequest(title, record, layout, minimized, access_level, plugin_path_list,
+                                          width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: bool = self.callback.data_record_form_view_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if not response:
             raise SapioUserCancelledException()
     
@@ -657,13 +815,19 @@ class CallbackUtil:
 
         request = TempTableSelectionRequest(builder.get_temporary_data_type(), msg, values,
                                             multi_select=multi_select)
-        response: list[FieldMap] | None = self.callback.show_temp_table_selection_dialog(request)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[FieldMap] | None = self.callback.show_temp_table_selection_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
     
     def record_selection_dialog(self, msg: str, fields: list[str], records: list[SapioRecord],
-                                multi_select: bool = True) -> list[FieldMap]:
+                                multi_select: bool = True) -> list[SapioRecord]:
         """
         Create a record selection dialog for a list of records for the user to choose from. Provided field names must
         match fields on the definition of the data type of the given records.
@@ -676,8 +840,10 @@ class CallbackUtil:
             they are provided in this list.
         :param records: The records to display as rows in the table.
         :param multi_select: Whether the user is able to select multiple records from the list.
-        :return: A list of field maps corresponding to the chosen input records.
+        :return: A list of the selected records.
         """
+        if not records:
+            raise SapioException("No records provided.")
         data_types: set[str] = {x.data_type_name for x in records}
         if len(data_types) > 1:
             raise SapioException("Multiple data type names encountered in records list for record table popup.")
@@ -707,7 +873,13 @@ class CallbackUtil:
 
         request = TempTableSelectionRequest(builder.get_temporary_data_type(), msg, field_map_list,
                                             multi_select=multi_select)
-        response: list[FieldMap] | None = self.callback.show_temp_table_selection_dialog(request)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[FieldMap] | None = self.callback.show_temp_table_selection_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         # Map the field maps in the response back to the record they come from, returning the chosen record instead of
@@ -774,7 +946,7 @@ class CallbackUtil:
 
         # If CustomReportCriteria was provided, it must be wrapped as a CustomReport.
         if isinstance(custom_search, CustomReportCriteria):
-            custom_search: CustomReport = CustomReport(False, None, custom_search)
+            custom_search: CustomReport = CustomReport(False, [], custom_search)
         # If a string was provided, locate the report criteria for the predefined search in the system matching this
         # name.
         if isinstance(custom_search, str):
@@ -783,7 +955,13 @@ class CallbackUtil:
         request = InputSelectionRequest(data_type, msg, search_types, only_key_fields, record_blacklist,
                                         record_whitelist, preselected_records, custom_search, scan_criteria,
                                         multi_select)
-        response: list[DataRecord] | None = self.callback.show_input_selection_dialog(request)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: list[DataRecord] | None = self.callback.show_input_selection_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return RecordHandler(self.user).wrap_models(response, wrapper_type)
@@ -807,14 +985,21 @@ class CallbackUtil:
             for field in additional_fields:
                 builder.add_field(field)
             temp_dt = builder.get_temporary_data_type()
-        request = ESigningRequestPojo(title, msg, show_comment, temp_dt)
-        response: ESigningResponsePojo | None = self.callback.show_esign_dialog(request)
+        request = ESigningRequestPojo(title, msg, show_comment, temp_dt,
+                                      width_in_pixels=self.width_pixels, width_percentage=self.width_percent)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            response: ESigningResponsePojo | None = self.callback.show_esign_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if response is None:
             raise SapioUserCancelledException()
         return response
 
     def request_file(self, title: str, exts: list[str] | None = None,
-                     show_image_editor: bool = False, show_camera_button: bool = False) -> (str, bytes):
+                     show_image_editor: bool = False, show_camera_button: bool = False) -> tuple[str, bytes]:
         """
         Request a single file from the user.
 
@@ -839,7 +1024,13 @@ class CallbackUtil:
                 return sink.consume_data(chunk, io_obj)
 
             request = FilePromptRequest(title, show_image_editor, ",".join(exts), show_camera_button)
-            file_path: str | None = self.callback.show_file_dialog(request, do_consume)
+            try:
+                self.user.timeout_seconds = self.timeout_seconds
+                file_path: str | None = self.callback.show_file_dialog(request, do_consume)
+            except ReadTimeout:
+                raise SapioDialogTimeoutException()
+            finally:
+                self.user.timeout_seconds = self._original_timeout
         if file_path is None:
             raise SapioUserCancelledException()
 
@@ -847,7 +1038,7 @@ class CallbackUtil:
         return file_path, sink.data
 
     def request_files(self, title: str, exts: list[str] | None = None,
-                      show_image_editor: bool = False, show_camera_button: bool = False):
+                      show_image_editor: bool = False, show_camera_button: bool = False) -> dict[str, bytes]:
         """
         Request multiple files from the user.
 
@@ -864,7 +1055,13 @@ class CallbackUtil:
             exts: list[str] = []
 
         request = MultiFilePromptRequest(title, show_image_editor, ",".join(exts), show_camera_button)
-        file_paths: list[str] | None = self.callback.show_multi_file_dialog(request)
+        try:
+            self.user.timeout_seconds = self.timeout_seconds
+            file_paths: list[str] | None = self.callback.show_multi_file_dialog(request)
+        except ReadTimeout:
+            raise SapioDialogTimeoutException()
+        finally:
+            self.user.timeout_seconds = self._original_timeout
         if not file_paths:
             raise SapioUserCancelledException()
 
@@ -878,7 +1075,7 @@ class CallbackUtil:
         return ret_dict
 
     @staticmethod
-    def __verify_file(file_path: str, file_bytes: bytes, allowed_extensions: list[str]):
+    def __verify_file(file_path: str, file_bytes: bytes, allowed_extensions: list[str]) -> None:
         """
         Verify that the provided file was read (i.e. the file path and file bytes aren't None or empty) and that it
         has the correct file extension. Raises a user error exception if something about the file is incorrect.
@@ -892,7 +1089,7 @@ class CallbackUtil:
         if len(allowed_extensions) != 0:
             matches: bool = False
             for ext in allowed_extensions:
-                if file_path.endswith("." + ext):
+                if file_path.endswith("." + ext.lstrip(".")):
                     matches = True
                     break
             if matches is False:
@@ -906,9 +1103,19 @@ class CallbackUtil:
         :param file_name: The name of the file.
         :param file_data: The data of the file, provided as either a string or as a bytes array.
         """
-        data = io.StringIO(file_data) if isinstance(file_data, str) else io.BytesIO(file_data)
+        data = io.BytesIO(file_data.encode() if isinstance(file_data, str) else file_data)
         self.callback.send_file(file_name, False, data)
 
+    def write_zip_file(self, zip_name: str, files: dict[str, str | bytes]) -> None:
+        """
+        Send a collection of files to the user in a zip file.
+
+        :param zip_name: The name of the zip file.
+        :param files: A dictionary of the files to add to the zip file.
+        """
+        data = io.BytesIO(FileUtil.zip_files(files))
+        self.callback.send_file(zip_name, False, data)
+
 
 class FieldModifier:
     """
@@ -956,7 +1163,7 @@ class FieldModifier:
         """
         field = copy_field(field)
         if self.prepend_data_type is True:
-            field._data_field_name = field.data_field_name + "." + field.data_field_name
+            field._data_field_name = field.data_type_name + "." + field.data_field_name
         if self.display_name is not None:
             field.display_name = self.display_name
         if self.required is not None:
