@@ -4,13 +4,14 @@ import urllib.parse
 
 from requests import Response
 from sapiopylib.rest.User import SapioUser
-from sapiopylib.rest.pojo.webhook.WebhookContext import SapioWebhookContext
+
+from sapiopycommons.general.aliases import UserIdentifier, AliasUtil
 
 
 # FR-46064 - Initial port of PyWebhookUtils to sapiopycommons.
 class FileBridge:
     @staticmethod
-    def read_file(context: SapioWebhookContext | SapioUser, bridge_name: str, file_path: str,
+    def read_file(context: UserIdentifier, bridge_name: str, file_path: str,
                   base64_decode: bool = True) -> bytes:
         """
         Read a file from FileBridge.
@@ -27,7 +28,7 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
+        user: SapioUser = AliasUtil.to_sapio_user(context)
         response = user.get(sub_path, params)
         user.raise_for_status(response)
 
@@ -37,7 +38,7 @@ class FileBridge:
         return ret_val
 
     @staticmethod
-    def write_file(context: SapioWebhookContext | SapioUser, bridge_name: str, file_path: str,
+    def write_file(context: UserIdentifier, bridge_name: str, file_path: str,
                    file_data: bytes | str) -> None:
         """
         Write a file to FileBridge.
@@ -53,13 +54,13 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
-        with io.StringIO(file_data) if isinstance(file_data, str) else io.BytesIO(file_data) as data_stream:
+        user: SapioUser = AliasUtil.to_sapio_user(context)
+        with io.BytesIO(file_data.encode() if isinstance(file_data, str) else file_data) as data_stream:
             response = user.post_data_stream(sub_path, params=params, data_stream=data_stream)
         user.raise_for_status(response)
 
     @staticmethod
-    def list_directory(context: SapioWebhookContext | SapioUser, bridge_name: str,
+    def list_directory(context: UserIdentifier, bridge_name: str,
                        file_path: str | None = "") -> list[str]:
         """
         List the contents of a FileBridge directory.
@@ -74,7 +75,7 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
+        user: SapioUser = AliasUtil.to_sapio_user(context)
         response: Response = user.get(sub_path, params=params)
         user.raise_for_status(response)
 
@@ -83,7 +84,7 @@ class FileBridge:
         return [urllib.parse.unquote(value)[path_length:] for value in response_body]
 
     @staticmethod
-    def create_directory(context: SapioWebhookContext | SapioUser, bridge_name: str, file_path: str) -> None:
+    def create_directory(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
         """
         Create a new directory in FileBridge.
 
@@ -97,12 +98,12 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
+        user: SapioUser = AliasUtil.to_sapio_user(context)
         response = user.post(sub_path, params=params)
         user.raise_for_status(response)
 
     @staticmethod
-    def delete_file(context: SapioWebhookContext | SapioUser, bridge_name: str, file_path: str) -> None:
+    def delete_file(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
         """
         Delete an existing file in FileBridge.
 
@@ -115,12 +116,12 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
+        user: SapioUser = AliasUtil.to_sapio_user(context)
         response = user.delete(sub_path, params=params)
         user.raise_for_status(response)
 
     @staticmethod
-    def delete_directory(context: SapioWebhookContext | SapioUser, bridge_name: str, file_path: str) -> None:
+    def delete_directory(context: UserIdentifier, bridge_name: str, file_path: str) -> None:
         """
         Delete an existing directory in FileBridge.
 
@@ -133,6 +134,6 @@ class FileBridge:
         params = {
             'Filepath': f"bridge://{bridge_name}/{file_path}"
         }
-        user: SapioUser = context if isinstance(context, SapioUser) else context.user
+        user: SapioUser = AliasUtil.to_sapio_user(context)
         response = user.delete(sub_path, params=params)
         user.raise_for_status(response)
